"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SdJwtVcService = void 0;
const core_1 = require("@sd-jwt/core");
const decode_1 = require("@sd-jwt/decode");
const tsyringe_1 = require("tsyringe");
const crypto_1 = require("../../crypto");
const utils_1 = require("../../utils");
const dids_1 = require("../dids");
const SdJwtVcError_1 = require("./SdJwtVcError");
const repository_1 = require("./repository");
/**
 * @internal
 */
let SdJwtVcService = class SdJwtVcService {
    constructor(sdJwtVcRepository) {
        this.sdJwtVcRepository = sdJwtVcRepository;
    }
    async sign(agentContext, options) {
        const { payload, disclosureFrame, hashingAlgorithm } = options;
        // default is sha-256
        if (hashingAlgorithm && hashingAlgorithm !== 'sha-256') {
            throw new SdJwtVcError_1.SdJwtVcError(`Unsupported hashing algorithm used: ${hashingAlgorithm}`);
        }
        const issuer = await this.extractKeyFromIssuer(agentContext, options.issuer);
        const holderBinding = await this.extractKeyFromHolderBinding(agentContext, options.holder);
        const header = {
            alg: issuer.alg,
            typ: 'vc+sd-jwt',
            kid: issuer.kid,
        };
        const sdJwtVc = new core_1.SdJwtVc({}, { disclosureFrame })
            .withHasher(this.hasher)
            .withSigner(this.signer(agentContext, issuer.key))
            .withSaltGenerator(agentContext.wallet.generateNonce)
            .withHeader(header)
            .withPayload(Object.assign({}, payload));
        // Add the `cnf` claim for the holder key binding
        sdJwtVc.addPayloadClaim('cnf', holderBinding.cnf);
        // Add `iss` claim
        sdJwtVc.addPayloadClaim('iss', issuer.iss);
        // Add the issued at (iat) claim
        sdJwtVc.addPayloadClaim('iat', Math.floor(new Date().getTime() / 1000));
        const compact = await sdJwtVc.toCompact();
        if (!sdJwtVc.signature) {
            throw new SdJwtVcError_1.SdJwtVcError('Invalid sd-jwt-vc state. Signature should have been set when calling `toCompact`.');
        }
        return {
            compact,
            prettyClaims: await sdJwtVc.getPrettyClaims(),
            header: sdJwtVc.header,
            payload: sdJwtVc.payload,
        };
    }
    fromCompact(compactSdJwtVc) {
        // NOTE: we use decodeSdJwtVc so we can make this method sync
        const { decodedPayload, header, signedPayload } = (0, decode_1.decodeSdJwtVc)(compactSdJwtVc, utils_1.Hasher.hash);
        return {
            compact: compactSdJwtVc,
            header: header,
            payload: signedPayload,
            prettyClaims: decodedPayload,
        };
    }
    async present(agentContext, { compactSdJwtVc, presentationFrame, verifierMetadata }) {
        const sdJwtVc = core_1.SdJwtVc.fromCompact(compactSdJwtVc).withHasher(this.hasher);
        const holder = await this.extractKeyFromHolderBinding(agentContext, this.parseHolderBindingFromCredential(sdJwtVc));
        // FIXME: we create the SD-JWT in two steps as the _sd_hash is currently not included in the SD-JWT library
        // so we add it ourselves, but for that we need the contents of the derived SD-JWT first
        const compactDerivedSdJwtVc = await sdJwtVc.present(presentationFrame === true ? undefined : presentationFrame);
        let sdAlg;
        try {
            sdAlg = sdJwtVc.getClaimInPayload('_sd_alg');
        }
        catch (error) {
            sdAlg = 'sha-256';
        }
        const header = {
            alg: holder.alg,
            typ: 'kb+jwt',
        };
        const payload = {
            iat: verifierMetadata.issuedAt,
            nonce: verifierMetadata.nonce,
            aud: verifierMetadata.audience,
            // FIXME: _sd_hash is missing. See
            // https://github.com/berendsliedrecht/sd-jwt-ts/issues/8
            _sd_hash: utils_1.TypedArrayEncoder.toBase64URL(utils_1.Hasher.hash(compactDerivedSdJwtVc, sdAlg)),
        };
        const compactKbJwt = await new core_1.KeyBinding({ header, payload })
            .withSigner(this.signer(agentContext, holder.key))
            .toCompact();
        return `${compactDerivedSdJwtVc}${compactKbJwt}`;
    }
    async verify(agentContext, { compactSdJwtVc, keyBinding, requiredClaimKeys }) {
        const sdJwtVc = core_1.SdJwtVc.fromCompact(compactSdJwtVc).withHasher(this.hasher);
        const issuer = await this.extractKeyFromIssuer(agentContext, this.parseIssuerFromCredential(sdJwtVc));
        const holder = await this.extractKeyFromHolderBinding(agentContext, this.parseHolderBindingFromCredential(sdJwtVc));
        // FIXME: we currently pass in the required keys in the verification method and based on the header.typ we
        // check if we need to use the issuer or holder key. Once better support in sd-jwt lib is available we can
        // update this.
        // See https://github.com/berendsliedrecht/sd-jwt-ts/pull/34
        // See https://github.com/berendsliedrecht/sd-jwt-ts/issues/15
        const verificationResult = await sdJwtVc.verify(this.verifier(agentContext, {
            issuer: issuer.key,
            holder: holder.key,
        }), requiredClaimKeys, holder.cnf);
        // If keyBinding is present, verify the key binding
        try {
            if (keyBinding) {
                if (!sdJwtVc.keyBinding || !sdJwtVc.keyBinding.payload) {
                    throw new SdJwtVcError_1.SdJwtVcError('Keybinding is required for verification of the sd-jwt-vc');
                }
                let sdAlg;
                try {
                    sdAlg = sdJwtVc.getClaimInPayload('_sd_alg');
                }
                catch (error) {
                    sdAlg = 'sha-256';
                }
                // FIXME: Calculate _sd_hash. can be removed once below is resolved
                // https://github.com/berendsliedrecht/sd-jwt-ts/issues/8
                const sdJwtParts = compactSdJwtVc.split('~');
                sdJwtParts.pop(); // remove kb-jwt
                const sdJwtWithoutKbJwt = `${sdJwtParts.join('~')}~`;
                const sdHash = utils_1.TypedArrayEncoder.toBase64URL(utils_1.Hasher.hash(sdJwtWithoutKbJwt, sdAlg));
                // Assert `aud` and `nonce` claims
                sdJwtVc.keyBinding.assertClaimInPayload('aud', keyBinding.audience);
                sdJwtVc.keyBinding.assertClaimInPayload('nonce', keyBinding.nonce);
                sdJwtVc.keyBinding.assertClaimInPayload('_sd_hash', sdHash);
            }
        }
        catch (error) {
            verificationResult.isKeyBindingValid = false;
            verificationResult.isValid = false;
        }
        return {
            verification: verificationResult,
            sdJwtVc: {
                payload: sdJwtVc.payload,
                header: sdJwtVc.header,
                compact: compactSdJwtVc,
                prettyClaims: await sdJwtVc.getPrettyClaims(),
            },
        };
    }
    async store(agentContext, compactSdJwtVc) {
        const sdJwtVcRecord = new repository_1.SdJwtVcRecord({
            compactSdJwtVc,
        });
        await this.sdJwtVcRepository.save(agentContext, sdJwtVcRecord);
        return sdJwtVcRecord;
    }
    async getById(agentContext, id) {
        return await this.sdJwtVcRepository.getById(agentContext, id);
    }
    async getAll(agentContext) {
        return await this.sdJwtVcRepository.getAll(agentContext);
    }
    async findByQuery(agentContext, query) {
        return await this.sdJwtVcRepository.findByQuery(agentContext, query);
    }
    async deleteById(agentContext, id) {
        await this.sdJwtVcRepository.deleteById(agentContext, id);
    }
    async update(agentContext, sdJwtVcRecord) {
        await this.sdJwtVcRepository.update(agentContext, sdJwtVcRecord);
    }
    async resolveDidUrl(agentContext, didUrl) {
        const didResolver = agentContext.dependencyManager.resolve(dids_1.DidResolverService);
        const didDocument = await didResolver.resolveDidDocument(agentContext, didUrl);
        return {
            verificationMethod: didDocument.dereferenceKey(didUrl, ['assertionMethod']),
            didDocument,
        };
    }
    get hasher() {
        return {
            algorithm: core_1.HasherAlgorithm.Sha256,
            hasher: utils_1.Hasher.hash,
        };
    }
    /**
     * @todo validate the JWT header (alg)
     */
    signer(agentContext, key) {
        return async (input) => agentContext.wallet.sign({ key, data: utils_1.TypedArrayEncoder.fromString(input) });
    }
    /**
     * @todo validate the JWT header (alg)
     */
    verifier(agentContext, verificationKeys) {
        return async ({ message, signature, publicKeyJwk, header }) => {
            const keyFromPublicKeyJwk = publicKeyJwk ? (0, crypto_1.getJwkFromJson)(publicKeyJwk).key : undefined;
            let key;
            if (header.typ === 'kb+jwt') {
                key = verificationKeys.holder;
            }
            else if (header.typ === 'vc+sd-jwt') {
                key = verificationKeys.issuer;
            }
            else {
                throw new SdJwtVcError_1.SdJwtVcError(`Unsupported JWT type '${header.typ}'`);
            }
            if (keyFromPublicKeyJwk && key.fingerprint !== keyFromPublicKeyJwk.fingerprint) {
                throw new SdJwtVcError_1.SdJwtVcError('The key used to verify the signature does not match the expected key');
            }
            return await agentContext.wallet.verify({
                signature: utils_1.Buffer.from(signature),
                key,
                data: utils_1.TypedArrayEncoder.fromString(message),
            });
        };
    }
    async extractKeyFromIssuer(agentContext, issuer) {
        if (issuer.method === 'did') {
            const parsedDid = (0, dids_1.parseDid)(issuer.didUrl);
            if (!parsedDid.fragment) {
                throw new SdJwtVcError_1.SdJwtVcError(`didUrl '${issuer.didUrl}' does not contain a '#'. Unable to derive key from did document`);
            }
            const { verificationMethod } = await this.resolveDidUrl(agentContext, issuer.didUrl);
            const key = (0, dids_1.getKeyFromVerificationMethod)(verificationMethod);
            const alg = (0, crypto_1.getJwkFromKey)(key).supportedSignatureAlgorithms[0];
            return {
                alg,
                key,
                iss: parsedDid.did,
                kid: `#${parsedDid.fragment}`,
            };
        }
        throw new SdJwtVcError_1.SdJwtVcError("Unsupported credential issuer. Only 'did' is supported at the moment.");
    }
    parseIssuerFromCredential(sdJwtVc) {
        const iss = sdJwtVc.getClaimInPayload('iss');
        if (iss.startsWith('did:')) {
            // If `did` is used, we require a relative KID to be present to identify
            // the key used by issuer to sign the sd-jwt-vc
            sdJwtVc.assertClaimInHeader('kid');
            const issuerKid = sdJwtVc.getClaimInHeader('kid');
            let didUrl;
            if (issuerKid.startsWith('#')) {
                didUrl = `${iss}${issuerKid}`;
            }
            else if (issuerKid.startsWith('did:')) {
                const didFromKid = (0, dids_1.parseDid)(issuerKid);
                if (didFromKid.did !== iss) {
                    throw new SdJwtVcError_1.SdJwtVcError(`kid in header is an absolute DID URL, but the did (${didFromKid.did}) does not match with the 'iss' did (${iss})`);
                }
                didUrl = issuerKid;
            }
            else {
                throw new SdJwtVcError_1.SdJwtVcError('Invalid issuer kid for did. Only absolute or relative (starting with #) did urls are supported.');
            }
            return {
                method: 'did',
                didUrl,
            };
        }
        throw new SdJwtVcError_1.SdJwtVcError("Unsupported 'iss' value. Only did is supported at the moment.");
    }
    parseHolderBindingFromCredential(sdJwtVc) {
        const cnf = sdJwtVc.getClaimInPayload('cnf');
        if (cnf.jwk) {
            return {
                method: 'jwk',
                jwk: cnf.jwk,
            };
        }
        else if (cnf.kid) {
            if (!cnf.kid.startsWith('did:') || !cnf.kid.includes('#')) {
                throw new SdJwtVcError_1.SdJwtVcError('Invalid holder kid for did. Only absolute KIDs for cnf are supported');
            }
            return {
                method: 'did',
                didUrl: cnf.kid,
            };
        }
        throw new SdJwtVcError_1.SdJwtVcError("Unsupported credential holder binding. Only 'did' and 'jwk' are supported at the moment.");
    }
    async extractKeyFromHolderBinding(agentContext, holder) {
        if (holder.method === 'did') {
            const parsedDid = (0, dids_1.parseDid)(holder.didUrl);
            if (!parsedDid.fragment) {
                throw new SdJwtVcError_1.SdJwtVcError(`didUrl '${holder.didUrl}' does not contain a '#'. Unable to derive key from did document`);
            }
            const { verificationMethod } = await this.resolveDidUrl(agentContext, holder.didUrl);
            const key = (0, dids_1.getKeyFromVerificationMethod)(verificationMethod);
            const alg = (0, crypto_1.getJwkFromKey)(key).supportedSignatureAlgorithms[0];
            return {
                alg,
                key,
                cnf: {
                    // We need to include the whole didUrl here, otherwise the verifier
                    // won't know which did it is associated with
                    kid: holder.didUrl,
                },
            };
        }
        else if (holder.method === 'jwk') {
            const jwk = holder.jwk instanceof crypto_1.Jwk ? holder.jwk : (0, crypto_1.getJwkFromJson)(holder.jwk);
            const key = jwk.key;
            const alg = jwk.supportedSignatureAlgorithms[0];
            return {
                alg,
                key,
                cnf: {
                    jwk: jwk.toJson(),
                },
            };
        }
        throw new SdJwtVcError_1.SdJwtVcError("Unsupported credential holder binding. Only 'did' and 'jwk' are supported at the moment.");
    }
};
SdJwtVcService = __decorate([
    (0, tsyringe_1.injectable)(),
    __metadata("design:paramtypes", [repository_1.SdJwtVcRepository])
], SdJwtVcService);
exports.SdJwtVcService = SdJwtVcService;
//# sourceMappingURL=SdJwtVcService.js.map